unit unTfLite;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface
{******************************************************************************
  MUST ALWAYS DISABLE OPTIMIZATION in release, because it causes memory access error
*******************************************************************************}
{$OPTIMIZATION OFF}

uses

{$IFnDEF FPC}
  Vcl.Dialogs, Graphics,
{$ELSE}
  LCLIntf, LCLType, LMessages, Graphics,
{$ENDIF}
  Windows,  messages, sysutils, variants, Classes,
  Math,
  unTFLiteResults;


  type
  TfLiteStatus = (kTfLiteOk, kTfLiteError, kTfLiteDelegateError);
  TfLiteType = (kTfLiteNoType = 0, kTfLiteFloat32 = 1, kTfLiteInt32 = 2,
    kTfLiteUInt8 = 3, kTfLiteInt64 = 4, kTfLiteString = 5, kTfLiteBool = 6,
    kTfLiteInt16 = 7, kTfLiteComplex64 = 8, kTfLiteInt8 = 9,
    kTfLiteFloat16 = 10, kTfLiteFloat64 = 11);

  Float32 = Single;

  TArray2d = array of array of Single;

  TTensorInfo = record
    name: string;
    numDims: Integer;
    datatype: string;
    byteSize: NativeUInt;
    subdims: array of Integer;
  end;

  TTFLiteInterpreterBase = class
  private
    fInputWidth: Integer;
    fInputHeight: Integer;
    fInputDim: Integer;
    fModelName: string;
    fLabelName: string;

    lstLabels: TStringList;
    fLabelNum: Integer;

    fVersion: string;
    fLibrary: HMODULE;
    fModel: Pointer;
    fInterpreterOptions: Pointer;
    fStatus: TfLiteStatus;
    fInputTensorCount, fOutputTensorCount, fNumDims: Int32;
    fInputTensor, fOutputTensor: Pointer;
    fInterpreter: Pointer;

    fTensorName: PAnsiChar;
    fTensorType: TfLiteType;
    fTensorInputByteSize: SIZE_T;

    fNumOutDims:    Int32;
    fTensorOutName: PAnsiChar;
    fTensorOutType: TfLiteType;
    fTensorOutByteSize: SIZE_T;

    fTensorByteSize: SIZE_T;

    procedure init(modelname: string);
    procedure readOutput(tensorNum: Integer; outputArray: Pointer);
    procedure verifyPredictInput(img: Graphics.TBitmap; outlist: TTFlstResults);
    procedure prepareInputInt(img: Graphics.TBitmap); virtual;
    procedure prepareInputFloat(img: Graphics.TBitmap; const normalize: Boolean); virtual;
  protected

    procedure prepareInputArrFloat(arr: TArray2d; const normalize: Boolean;
              const normalizeBase: Single); virtual;
    procedure prepareInput(img: Graphics.TBitmap; const datatype: TfLiteType; const normalize: Boolean = False); virtual;
  public
    property inputWidth: Integer read fInputWidth;
    property inputHeight: integer read fInputHeight;
    property inputDim: Integer read fInputDim;
    property labelNum: integer read fLabelNum;
    property modelname: string read fModelName;
    property version: string read fVersion;

    property inpTensorsCount: integer read fInputTensorCount;
    property outTensorsCount: Integer read fOutputTensorCount;
    constructor create(modelfile: string; labelfile: string = 'labelmap.txt');
    destructor destroy(); override;

    function getInpTensorInfo(ind: Integer): TTensorInfo;
    function getOutTensorInfo(ind: Integer): TTensorInfo;
    procedure doPrediction(img: Graphics.TBitmap; outlist: TTFlstResults);    virtual;
  end;

  TTFLiteMobilenetv2 = class(TTFLiteInterpreterBase)
  private
    fMinScore: Single;
  protected
  public
    property minScore: Single read fMinScore write fMinScore;
    constructor create(modelfile: string; labelname: string = 'labelmap.txt');
    procedure doPrediction(img: Graphics.TBitmap; outlist: TTFlstResults);  override;
  end;


  TTFLiteSSDMobilenet = class(TTFLiteInterpreterBase)
  private
    fMinScore: Single;
  protected
  public
    property minScore: Single read fMinScore write fMinScore;
    constructor create(modelfile: string; labelname: string = 'labelmap.txt');
    procedure doPrediction(img: Graphics.TBitmap; outlist: TTFlstResults);  override;
  end;


  TTFLiteSSDMobilenetFloat = class(TTFLiteSSDMobilenet)
  private
  protected
  public
    constructor create(modelfile: string; labelname: string = 'labelmap.txt');
    procedure doPrediction(img: Graphics.TBitmap; outlist: TTFlstResults);  override;
  end;




  TTFLiteCenternetMobilenet = class(TTFLiteSSDMobilenet)
  private
     outputInd: array[0..3] of Integer;
     tensorsIndexFound: Boolean;
  protected
  public
    constructor create(modelfile: string; labelname: string = 'labelmap.txt');
    procedure doPrediction(img: Graphics.TBitmap; outlist: TTFlstResults);  override;
  end;

{*****************************************************************************}
implementation

uses
{$IFnDEF FPC}
  System.TypInfo;
{$ELSE}
  typinfo;
{$ENDIF}


const
{$IFDEF WIN32}
  LibraryName = 'tflite32.dll';
{$ENDIF}
{$IFDEF WIN64}
  LibraryName = 'tflite64.dll';
{$ENDIF}

  INPUT_DIM = 1024;




function TfLiteModelCreateFromFile(const model_path: PAnsiChar): Pointer;
  stdcall; external LibraryName;

function TfLiteInterpreterCreate(model: Pointer; optional_options: Pointer)
  : Pointer; stdcall; external LibraryName;

function TfLiteInterpreterOptionsCreate(): Pointer; stdcall;
  external LibraryName;

function TfLiteInterpreterAllocateTensors(interpreter: Pointer): TfLiteStatus;
  stdcall; external LibraryName;

function TfLiteInterpreterGetInputTensor(interpreter: Pointer;
  input_index: Int32): Pointer; stdcall; external LibraryName;

function TfLiteInterpreterGetOutputTensor(interpreter: Pointer;
  output_index: Int32): Pointer; stdcall; external LibraryName;

function TfLiteTensorNumDims(tensor: Pointer): Int32; stdcall;
  external LibraryName;

// Returns the length of the tensor in the "dim_index" dimension.
// REQUIRES: 0 <= dim_index < TFLiteTensorNumDims(tensor)
function TfLiteTensorDim(tensor: Pointer; dim_index: Int32): int32; stdcall;
  external LibraryName;
function TfLiteTensorName(tensor: Pointer): PAnsiChar; stdcall;
  external LibraryName;

function TfLiteTensorType(tensor: Pointer): TfLiteType; stdcall;
  external LibraryName;

function TfLiteTensorByteSize(tensor: Pointer): SIZE_T; stdcall;
  external LibraryName;

function TfLiteInterpreterResizeInputTensor(interpreter: Pointer;
  input_index: Int32; input_dims: PInteger; input_dims_size: Int32)
  : TfLiteStatus; stdcall; external LibraryName;

function TfLiteInterpreterGetInputTensorCount(interpreter: Pointer): Int32;
  stdcall; external LibraryName;

function TfLiteInterpreterGetOutputTensorCount(interpreter: Pointer): Int32;
  stdcall; external LibraryName;

function TfLiteTensorCopyFromBuffer(tensor: Pointer; input_data: Pointer;
  input_data_size: SIZE_T): TfLiteStatus; stdcall; external LibraryName;

function TfLiteTensorCopyToBuffer(output_tensor: Pointer; output_data: Pointer;
  output_data_size: SIZE_T): TfLiteStatus; stdcall; external LibraryName;

procedure TfLiteInterpreterOptionsSetNumThreads(options: Pointer;
  num_threads: Int32); stdcall; external LibraryName;

procedure TfLiteInterpreterOptionsDelete(options: Pointer); stdcall;
  external LibraryName;

procedure TfLiteModelDelete(model: Pointer); stdcall; external LibraryName;

function TfLiteInterpreterInvoke(interpreter: Pointer): TfLiteStatus; stdcall;
  external LibraryName;

// TfLiteVersion returns a string describing version information of the
// TensorFlow Lite library. TensorFlow Lite uses semantic versioning.
function  TfLiteVersion(): PAnsiChar; stdcall; external LibraryName;



procedure raiseerror(msg: string);
begin
  raise Exception.Create(msg);
end;





{******************************************************************************}
{******************** TTFLiteInterpreterBase **********************************}
{******************************************************************************}

const
  // dimensions depend from neural network type
  LABEL_NUM = 1;

constructor TTFLiteInterpreterBase.create(modelfile: string; labelfile: string);
begin
  inherited create;
  fLabelName:=labelfile;
  init(modelfile);
end;

destructor TTFLiteInterpreterBase.Destroy;
begin
  lstLabels.Free;
  if fLibrary<>0 then
  begin
    FreeLibrary(fLibrary);
  end;
  inherited Destroy;
end;


procedure TTFLiteInterpreterBase.init(modelname: string);
begin
  if not(FileExists(modelname)) then
    raiseerror('The model file name not exists: '+modelname);

  lstLabels:=TStringList.Create;
  fLabelNum:=LABEL_NUM;
  lstLabels.Clear;
  if fLabelName<>'' then
      try
        lstLabels.LoadFromFile(fLabelName);
        fLabelNum:=lstLabels.Count;
      except
        on E: Exception do
          raiseerror('Error: loading labels '+E.Message);
      end;

  fInputWidth:=INPUT_DIM;
  fInputHeight:=INPUT_DIM;

  fLibrary:=0;
  fLibrary := LoadLibrary(LibraryName);

  if fLibrary = 0 then
  begin
    raiseerror('Error: Load tensorflow lite library ' + LibraryName + ' - ' +
      SysErrorMessage(GetLastError));
    Exit;
  end;


  fModel := TfLiteModelCreateFromFile(PAnsiChar(ansistring(modelname)));

  if fModel = nil then
  begin
    raiseerror('Error: Create model from file - ' +
          SysErrorMessage(GetLastError));
    Exit;
  end;

  fInterpreterOptions := TfLiteInterpreterOptionsCreate;

  if fInterpreterOptions = nil then
  begin
    raiseerror('Error: Create model options - ' +
          SysErrorMessage(GetLastError));
    Exit;
  end;
  TfLiteInterpreterOptionsSetNumThreads(fInterpreterOptions, 4);

  fInterpreter := TfLiteInterpreterCreate(fModel, fInterpreterOptions);
  if fInterpreter = nil then
  begin
    raiseerror('Error: Create interpreter - ' +
          SysErrorMessage(GetLastError));
    Exit;
  end;

  TfLiteInterpreterOptionsDelete(fInterpreterOptions);
  TfLiteModelDelete(fModel);

  fStatus := TfLiteInterpreterAllocateTensors(fInterpreter);

  fInputTensorCount := TfLiteInterpreterGetInputTensorCount(fInterpreter);
  fOutputTensorCount := TfLiteInterpreterGetOutputTensorCount(fInterpreter);

  fInputTensor := TfLiteInterpreterGetInputTensor(fInterpreter, 0);
  if fInputTensor = nil then
  begin
    raiseerror('Error: Create input tensor - ' +
          SysErrorMessage(GetLastError));
    Exit;
  end;
  fOutputTensor := TfLiteInterpreterGetOutputTensor(fInterpreter, 0);
  if fOutputTensor = nil then
  begin
    raiseerror('Error: Create output tensor - ' +
          SysErrorMessage(GetLastError));
    Exit;
  end;
  // input tensor info
  fNumDims := TfLiteTensorNumDims(fInputTensor);
  fTensorName := TfLiteTensorName(fInputTensor);
  fTensorType := TfLiteTensorType(fInputTensor);
  fTensorInputByteSize := TfLiteTensorByteSize(fInputTensor);

  // output tensor info
  fNumOutDims := TfLiteTensorNumDims(fOutputTensor);
  fTensorOutName := TfLiteTensorName(fOutputTensor);
  fTensorOutType := TfLiteTensorType(fOutputTensor);
  fTensorOutByteSize := TfLiteTensorByteSize(fOutputTensor);



  fVersion:=string(TfLiteVersion());

end;

function TTFLiteInterpreterBase.getInpTensorInfo(ind: Integer): TTensorInfo;
var
  i: Integer;
  inputtensor: Pointer;
  ty: TfLiteType;
begin
  inputTensor := TfLiteInterpreterGetInputTensor(fInterpreter, ind);
  Result.byteSize := TfLiteTensorByteSize(inputtensor);
  Result.numDims := TfLiteTensorNumDims(inputtensor);
  Result.name := TfLiteTensorName(inputtensor);
  ty:=  TfLiteTensorType(inputtensor);
  Result.datatype := GetEnumName(TypeInfo(TfLiteType), PByte(@(ty))^);
  SetLength(Result.subdims, Result.numDims);
  for i:=0 to Result.numDims-1 do
  begin
     Result.subdims[i]:= TfLiteTensorDim(inputtensor, i);
  end;
end;

function TTFLiteInterpreterBase.getOutTensorInfo(ind: Integer): TTensorInfo;
var
  i: Integer;
  outputtensor: Pointer;
  ty: TfLiteType;
begin
  OutputTensor := TfLiteInterpreterGetOutputTensor(fInterpreter, ind);
  Result.byteSize := TfLiteTensorByteSize(OutputTensor);
  Result.numDims := TfLiteTensorNumDims(OutputTensor);
  Result.name := TfLiteTensorName(OutputTensor);
  ty:=  TfLiteTensorType(OutputTensor);
  Result.datatype := GetEnumName(TypeInfo(TfLiteType), PByte(@(ty))^);
  SetLength(Result.subdims, Result.numDims);
  for i:=0 to Result.numDims-1 do
  begin
     Result.subdims[i]:= TfLiteTensorDim(outputtensor, i);
  end;
end;




procedure TTFLiteInterpreterBase.prepareInputFloat(img: Graphics.TBitmap; const normalize: Boolean);
const
  PIXELCOUNT = INPUT_DIM * INPUT_DIM;
type
  PRGBArray = ^TRGBArray;
  TRGBArray = array [0 .. PIXELCOUNT - 1] of TRGBTriple;

  PInputArray = ^TInputArray;
  TInputArray = array [0 .. PIXELCOUNT - 1] of array [0 .. 3 - 1] of Float32;
var
  fInputArray: PInputArray;
  fColors: PRGBArray;

  X, Y: Integer;

begin
  GetMem(fInputArray,fTensorInputByteSize);

  try
  // with "if" out of cycle there is a gain in speed
  if normalize  then
  begin
    for Y := 0 to img.Height - 1 do
    begin
      fColors := PRGBArray(img.ScanLine[Y]);
      for X := 0 to img.Width - 1 do
      begin
        fInputArray[X + (Y * img.Width)][0] :=
          (fColors[X].rgbtRed - 127.5) / 127.5;
        fInputArray[X + (Y * img.Width)][1] :=
          (fColors[X].rgbtGreen - 127.5) / 127.5;
        fInputArray[X + (Y * img.Width)][2] :=
          (fColors[X].rgbtBlue -  127.5) / 127.5;
      end;
    end;
  end;
  if not(normalize) then
  begin
    for Y := 0 to img.Height - 1 do
    begin
      fColors := PRGBArray(img.ScanLine[Y]);
      for X := 0 to img.Width - 1 do
      begin
        fInputArray[X + (Y * img.Width)][0] :=
          fColors[X].rgbtRed;
        fInputArray[X + (Y * img.Width)][1] :=
          fColors[X].rgbtGreen;
        fInputArray[X + (Y * img.Width)][2] :=
          fColors[X].rgbtBlue;
      end;
    end;

  end;
    fStatus := TfLiteTensorCopyFromBuffer(fInputTensor, fInputArray,
      fTensorInputByteSize);
 finally
    FreeMem(fInputArray, fTensorInputByteSize);
  end;

  if fStatus <> kTfLiteOk then
  begin
    raiseerror('Error: Loading input tensor - ' +
          IntToStr(Ord(fStatus)));
  end;


end;

procedure TTFLiteInterpreterBase.prepareInputArrFloat(arr: TArray2d; const normalize: Boolean;
           const normalizeBase: Single);
const
  DATACOUNT = INPUT_DIM * INPUT_DIM ;
type
  PInputArray = ^TInputArray;
  TInputArray = array [0 .. DATACOUNT - 1]  of Float32;
var
  fInputArray: PInputArray;

  X, Y: Integer;

  outind: Integer;
begin
  if (Length(arr) > INPUT_DIM) or (Length(arr[0])>INPUT_DIM) then
      raiseerror('Input array size can be 1024x1024 max ');
  GetMem(fInputArray,fTensorInputByteSize);

  try
  outind:=0;
  // with "if" out of cycle there is a gain in speed
  if normalize  then
  begin
    for X := Low(arr) to High(arr)  do
    begin
      for Y := Low(arr[0]) to High(arr[0])  do
      begin
        fInputArray[outind] :=
          (arr[X,Y] - normalizeBase) / normalizeBase;
        inc(outind);
      end;
    end;
  end;
  if not(normalize) then
  begin
    for X := Low(arr) to High(arr)  do
    begin
      for Y := Low(arr[0]) to High(arr[0])  do
      begin
        fInputArray[outind] :=  arr[X,Y];
        inc(outind);
      end;
    end;

  end;
    fStatus := TfLiteTensorCopyFromBuffer(fInputTensor, fInputArray,
      fTensorInputByteSize);
 finally
    FreeMem(fInputArray, fTensorInputByteSize);
  end;

  if fStatus <> kTfLiteOk then
  begin
    raiseerror('Error: Loading input tensor - ' +
          IntToStr(Ord(fStatus)));
  end;
end;



procedure TTFLiteInterpreterBase.prepareInputInt(img: Graphics.TBitmap);
 // A standard procedure to prepare the input image
 // in format required  from neural network. The maximum image
 // dimension are 1024x1024, the network input type if Uint8.
 // If dimensions are greater , or type is different,
 // the procedure has to be overloaded
const
  PIXELCOUNT = INPUT_DIM * INPUT_DIM;
type
  PRGBArray = ^TRGBArray;
  TRGBArray = array [0 .. PIXELCOUNT - 1] of TRGBTriple;

  PInputArray = ^TInputArray;
  TInputArray = array [0 .. INPUT_DIM * INPUT_DIM - 1] of array [0 .. 3 - 1] of UInt8;
var
  fInputArray: PInputArray;
  fColors: PRGBArray;

  X, Y: Integer;
begin
  GetMem(fInputArray,fTensorInputByteSize);
  try
    for Y := 0 to img.Height - 1 do
    begin
      fColors := PRGBArray(img.ScanLine[Y]);

      for X := 0 to img.Width - 1 do
      begin
            fInputArray[X + (Y * img.Width)][0] :=
              fColors[X].rgbtRed;
            fInputArray[X + (Y * img.Width)][1] :=
              fColors[X].rgbtGreen;
            fInputArray[X + (Y * img.Width)][2] :=
              fColors[X].rgbtBlue;
      end;
    end;

    fStatus := TfLiteTensorCopyFromBuffer(fInputTensor, fInputArray,
      fTensorInputByteSize);
 finally
    FreeMem(fInputArray, fTensorInputByteSize);
  end;

  if fStatus <> kTfLiteOk then
  begin
    raiseerror('Error: Loading input tensor - ' +
          IntToStr(Ord(fStatus)));
  end;

end;

procedure TTFLiteInterpreterBase.prepareInput(img: Graphics.TBitmap; const datatype: TfLiteType; const normalize: Boolean);
begin
    if (img.Height > INPUT_DIM) or (img.Width>INPUT_DIM) then
      raiseerror('Input image size can be 1024x1024 max ');
    if datatype=TfLiteType.kTfLiteFloat32 then
    begin
        prepareInputFloat(img, normalize);
        Exit;
    end ;
    if datatype=TfLiteType.kTfLiteUInt8 then
    begin
        if normalize=false then
          prepareInputInt(img)
        else
          raiseerror('Input data of type Uint8 cannot be normalized');
        Exit;
    end;

    raiseerror('Input data can be only of type Uint8 or Float');
end;


procedure TTFLiteInterpreterBase.verifyPredictInput(img: Graphics.TBitmap; outlist: TTFlstResults);
begin
  if not(Assigned(img)) then
  begin
    raiseerror('Error in prediction: input image not assigned');
  end;
  if not(Assigned(outlist)) then
  begin
    raiseerror('Error in prediction: output results list not assigned');
  end;

  outlist.Clear;

end;

procedure TTFLiteInterpreterBase.doPrediction(img: Graphics.TBitmap; outlist: TTFlstResults);
begin
  //nop
end;


procedure TTFLiteInterpreterBase.readOutput(tensorNum: Integer; outputArray: Pointer);
var
  internalDim: Int32;
begin
  fOutputTensor := TfLiteInterpreterGetOutputTensor(fInterpreter, tensorNum);
  fTensorByteSize := TfLiteTensorByteSize(fOutputTensor);
  fNumOutDims := TfLiteTensorNumDims(fOutputTensor);
  fTensorOutName := TfLiteTensorName(fOutputTensor);
  fTensorOutType := TfLiteTensorType(fOutputTensor);
  internalDim := TfLiteTensorDim(fOutputTensor, 0);

  if fOutputTensor <> nil then
  begin
    fStatus := TfLiteTensorCopyToBuffer(fOutputTensor, outputArray, fTensorByteSize);
    if fStatus <> kTfLiteOk then
    begin
      raiseerror('Error: copying tensor ' + IntToStr(tensorNum) + ' - ' +
            IntToStr(Ord(fStatus)));
    end;
  end;

end;




{******************************************************************************}
{******************** TTFLiteSSDMobilenet  ************************************}
{******************************************************************************}

constructor TTFLiteSSDMobilenet.create(modelfile: string; labelname: string);
begin
  inherited create(modelfile, labelname);

  fInputWidth:= 300;
  fInputHeight:=300;
  fInputDim:=300;
  fMinScore:=0.5;
end;



procedure TTFLiteSSDMobilenet.doPrediction(img: Graphics.TBitmap; outlist: TTFlstResults);
const
  RES_NUM = 10;
var
  i: Integer;

  fOutputLocations: array [0 .. RES_NUM - 1] of array [0 .. 4 - 1] of Float32;
  fOutputClasses: array [0 .. RES_NUM - 1] of Float32;
  fOutputScores: array [0 .. RES_NUM - 1] of Float32;
//  fNumDetections: array [0 .. 1 - 1] of Float32;

  fValue: Extended;

  tfres: TTFBoxResult;
begin
  verifyPredictInput(img, outlist);


  prepareInput(img, kTfLiteUInt8);


  fStatus := TfLiteInterpreterInvoke(fInterpreter);

  if fStatus <> kTfLiteOk then
  begin
    raiseerror('Error: Invoking interpreter - ' +
          IntToStr(Ord(fStatus)));
  end;

  // Output is of 4 tensors:
  { 0	Locations	Multidimensional array of [10][4] floating point values between 0 and 1, the inner arrays representing bounding boxes in the form [top, left, bottom, right]
    1	Classes	Array of 10 integers (output as floating point values) each indicating the index of a class label from the labels file
    2	Scores	Array of 10 floating point values between 0 and 1 representing probability that a class was detected
    3	Number and detections }
  // index 0 : Locations
  readOutput(0, @fOutputLocations);

  // index 1 : Classes
  readOutput(1, @fOutputClasses);

  // index 2:  Scores
  readOutput(2, @fOutputScores);

  // index 3:  Number and detections
  // ...



  // There will always be 10 objects detected
  for i := 0 to Length(fOutputLocations[0]) - 1 do
  begin
    fValue := fOutputScores[i];

    if fValue >= fMinScore then
    begin
      tfres:=TTFBoxResult.Create;
      with tfres do
      begin
        classid:= Round(foutputClasses[i]);
        score :=fValue;
        posrect := Rect(Round(fOutputLocations[i][1] * fInputDim),
                        Round(fOutputLocations[i][0] * fInputDim),
                        Round(fOutputLocations[i][3] * fInputDim),
                        Round(fOutputLocations[i][2] * fInputDim)  );


      // SSD Mobilenet V1 Model assumes class 0 is background class
      // in label file and class labels start from 1 to number_of_classes+1,
      // while outputClasses correspond to class index from 0 to number_of_classes
      try
        labelname:=lstLabels.Strings[Round(fOutputClasses[i]) + 1];
      except
        labelname:='???';
      end;
      end;
      outlist.Add(tfres);
    end;
  end;

end;
{******************************************************************************}
{******************** TFLiteSSDMobilenetFloat  ********************************}
{******************************************************************************}


constructor TTFLiteSSDMobilenetFloat.create(modelfile: string; labelname: string);
begin
  inherited create(modelfile, labelname);

  fInputWidth:= 300;
  fInputHeight:=300;
  fInputDim:=300;
  fMinScore:=0.5;
end;



procedure TTFLiteSSDMobilenetFloat.doPrediction(img: Graphics.TBitmap; outlist: TTFlstResults);
const
  RES_NUM = 10;
var
  i: Integer;

  fOutputLocations: array [0 .. RES_NUM - 1] of array [0 .. 4 - 1] of Float32;
  fOutputClasses: array [0 .. RES_NUM - 1] of Float32;
  fOutputScores: array [0 .. RES_NUM - 1] of Float32;
//  fNumDetections: array [0 .. 1 - 1] of Float32;

  fValue: Extended;

  tfres: TTFBoxResult;
begin
  verifyPredictInput(img, outlist);


  prepareInput(img, kTfLiteFloat32, True);


  fStatus := TfLiteInterpreterInvoke(fInterpreter);

  if fStatus <> kTfLiteOk then
  begin
    raiseerror('Error: Invoking interpreter - ' +
          IntToStr(Ord(fStatus)));
  end;

  // Output is of 4 tensors:
  { 0	Locations	Multidimensional array of [10][4] floating point values between 0 and 1, the inner arrays representing bounding boxes in the form [top, left, bottom, right]
    1	Classes	Array of 10 integers (output as floating point values) each indicating the index of a class label from the labels file
    2	Scores	Array of 10 floating point values between 0 and 1 representing probability that a class was detected
    3	Number and detections }
  // index 0:  Scores
  readOutput(0, @fOutputScores);
  // index 1 : Locations
  readOutput(1, @fOutputLocations);

  // index 3 : Classes
  readOutput(3, @fOutputClasses);


  // index 2:  Number and detections
  // ...



  // There will always be 10 objects detected
  for i := 0 to Length(fOutputLocations[0]) - 1 do
  begin
    fValue := fOutputScores[i];

    if fValue >= fMinScore then
    begin
      tfres:=TTFBoxResult.Create;
      with tfres do
      begin
        classid:= Round(foutputClasses[i]);
        score :=fValue;
        posrect := Rect(Round(fOutputLocations[i][1] * fInputDim),
                        Round(fOutputLocations[i][0] * fInputDim),
                        Round(fOutputLocations[i][3] * fInputDim),
                        Round(fOutputLocations[i][2] * fInputDim)  );


      // SSD Mobilenet V1 Model assumes class 0 is background class
      // in label file and class labels start from 1 to number_of_classes+1,
      // while outputClasses correspond to class index from 0 to number_of_classes
      try
        labelname:=lstLabels.Strings[Round(fOutputClasses[i])];
      except
        labelname:='???';
      end;
      end;
      outlist.Add(tfres);
    end;
  end;

end;



{******************************************************************************}
{******************** TTFLiteMobilenetv2 **************************************}
{******************************************************************************}

constructor TTFLiteMobilenetv2.create(modelfile: string; labelname: string);
begin
  inherited create(modelfile, labelname);

  fInputWidth:=224;
  fInputHeight:=224;
  fInputDim:=224;
  fMinScore:=0.5;
  // output number = num of classes
  fNumOutDims := TfLiteTensorDim(fOutputTensor, 1);
end;




procedure TTFLiteMobilenetv2.doPrediction(img: Graphics.TBitmap; outlist: TTFlstResults);
const
  RES_NUM = 10;
var
  i: Integer;
  nrresults: Integer;

  fOutputScores: array [0 .. RES_NUM - 1] of Float32;

  fValue: Extended;

  tfres: TTFResult;
begin
  verifyPredictInput(img, outlist);


  prepareInput(img, kTfLiteFloat32, True);


  fStatus := TfLiteInterpreterInvoke(fInterpreter);

  if fStatus <> kTfLiteOk then
  begin
    raiseerror('Error: Invoking interpreter - ' +
          IntToStr(Ord(fStatus)));
  end;

  // Output is  1 tensor:
  { 0	Scores	Array of nrresults floating point values between 0 and 1 representing probability that a class was detected}

  // index 0: Scores
  readOutput(0, @fOutputScores);

  // output tensor nr. 0 has "nrresult" columns (index 1)
  nrresults:=TfLiteTensorDim(fOutputTensor, 1);

  for i := 0 to nrresults - 1 do
  begin
    fValue:=RoundTo(fOutputScores[i], -4);

    if fValue >= fMinScore then
    begin
      tfres:=TTFResult.Create;
      with tfres do
      begin
        classid:=i;
        score :=fValue;
        labelname:=lstLabels.Strings[i];
      end;
      outlist.Add(tfres);
    end;
  end;

end;

{******************************************************************************}
{******************** TTFLiteCenternetMobilenet *******************************}
{******************************************************************************}

constructor TTFLiteCenternetMobilenet.create(modelfile: string; labelname: string);
begin
  inherited create(modelfile, labelname);

  fInputWidth:= 320;
  fInputHeight:=320;
  fInputDim:=320;
  fMinScore:=0.5;

  outputInd[0]:=-1;
  outputInd[1]:=-1;
  outputInd[2]:=-1;
  outputInd[3]:=-1;

  tensorsIndexFound:=False;
end;


procedure TTFLiteCenternetMobilenet.doPrediction(img: Graphics.TBitmap; outlist: TTFlstResults);
const
  RES_NUM = 10;
var
  i, j: Integer;

  fOutputLocations: array [0 .. RES_NUM - 1] of array [0 .. 4 - 1] of Float32;
  fOutputClasses: array [0 .. RES_NUM - 1] of Float32;
  fOutputScores: array [0 .. RES_NUM - 1] of Float32;
  fNumDetections: array [0 .. 1 - 1] of Float32;

  fValue: Extended;

  tfres: TTFBoxResult;


  procedure searchTensors();
  var
    i,j: Integer;
  begin
        for i:=0 to 2 do
        begin
           fOutputLocations[i][0]:=-1.0;
           fOutputLocations[i][1]:=-1.0;
           fOutputLocations[i][2]:=-1.0;
           fOutputLocations[i][3]:=-1.0;
        end;

        // Search Locations tensor
        for i:=0 to 3 do
        begin
          readOutput(i, @fOutputLocations);
          // elements 9 and 10 have to be different from -1
          if (fOutputLocations[2][2]<>-1.0) and  (fOutputLocations[2][3]<>-1.0) then
          begin
            outputInd[0]:=i;  // save the index of tensor corresponding to locations
            Break;
          end;
        end;


        //  Search Scores tensor
        for i:=0 to 3 do
        begin
          if i=outputInd[0] then Continue;
          readOutput(i, @fOutputScores);
          // first element must be between 0 and 1
          if ((fOutputScores[0]>0) and (fOutputScores[0]<=1)) then
          begin
            outputInd[1]:=i;
            Break;
          end;
        end;


        // Number of detections tensor
        // it's always 10
        //readOutput(2, @fNumDetections);

        // Search  Classes tensor
        for i:=0 to 3 do
        begin
          if (i=outputInd[0]) or (i=outputInd[1]) then Continue;
          readOutput(i, @fOutputClasses);
          // every element must be integer
          try
            for j:=Low(fOutputClasses) to High(fOutputClasses) do
            begin
                if not(fOutputClasses[j]=trunc(fOutputClasses[j])) then
                begin
                  Continue;
                end;
            end;
          except
            Continue;
          end;
          outputInd[2]:=i;
          Break;
        end;

        tensorsIndexFound:=True;
  end;



begin
  verifyPredictInput(img, outlist);

  prepareInput(img, kTfLiteFloat32, False);

  fStatus := TfLiteInterpreterInvoke(fInterpreter);


  if fStatus <> kTfLiteOk then
  begin
    raiseerror('Error: Invoking interpreter - ' +
          IntToStr(Ord(fStatus)));
  end;

  // Output is of 4 tensors:
  { -	Locations	Multidimensional array of [10][4] floating point values between 0 and 1, the inner arrays representing bounding boxes in the form [top, left, bottom, right]
    -	Classes	Array of 10 integers (output as floating point values) each indicating the index of a class label from the labels file
    -	Scores	Array of 10 floating point values between 0 and 1 representing probability that a class was detected
    -	Number detections , always 10
    NOTE: the index of the outputs tensors may vary in different tflite files. So it's implemented a strategy to find
    the correct index for every tensor
  }

  if not(tensorsIndexFound) then
  begin
    searchTensors;
  end else
  begin
    readOutput(outputind[0], @fOutputLocations);
    readOutput(outputind[1], @fOutputScores);
    readOutput(outputind[2], @fOutputClasses);
  end;


  // There will always be 10 objects detected
  for i := 0 to Length(fOutputLocations[0]) - 1 do
  begin
    fValue := fOutputScores[i];

    if fValue >= fMinScore then
    begin
      tfres:=TTFBoxResult.Create;
      with tfres do
      begin
        classid:= Round(foutputClasses[i]);
        score :=fValue;
        posrect := Rect(Round(fOutputLocations[i][1] * fInputWidth),
                        Round(fOutputLocations[i][0] * fInputHeight),
                        Round(fOutputLocations[i][3] * fInputWidth),
                        Round(fOutputLocations[i][2] * fInputHeight)  );

      try
        labelname:=lstLabels.Strings[classid];
      except
        labelname:='???';
      end;
      end;
      outlist.Add(tfres);
    end;
  end;

end;




end.
