unit unTFLiteResults;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface
uses
{$IFnDEF FPC}
  Graphics, Vcl.Dialogs, System.Generics.Collections, Windows,
{$ELSE}
  LCLIntf, LCLType, LMessages, Generics.Collections, Graphics,
{$ENDIF}
  messages, sysutils, variants, Classes,
  Math;


type
  TTFResult = class
    classid:   Integer;
    labelname: string;
    score:     Single;
    id:        Integer;
  end;
  TTFBoxResult =  class(TTFResult)
    posrect:   TRect;
  end;

  TTFlstResults  = class(TObjectList<TTFResult>)
  public
    procedure drawBoxes(bmp: Graphics.TBitmap; id_zero: Boolean = True);
    procedure scale(scaleWidth: Single; scaleHeight: Single);
  end;

implementation
const
  TClassColors: array[0..7] of TColor =  (clRed, clBlue, clGreen, clYellow, clLtGray, clOlive,
  {$IFnDEF FPC}
      clWebCyan, clWebPink
  {$Else}
      clNavy, clPurple
  {$ENDIF}
  );


procedure TTFlstResults.drawBoxes(bmp: Graphics.TBitmap; id_zero: Boolean);
var
  tfres: TTFResult;
  cap: string;
  ltxt: Integer;
begin
    // draw boxes of objects detected by neural network
    bmp.Canvas.Font.Name:='Arial';
    bmp.Canvas.Font.Size:=10;
    bmp.Canvas.Pen.Style:=psDot;
    bmp.Canvas.Pen.Width:=2;
    for tfres in self do
    begin
        if  ((id_zero = False) and (tfres.id=0)) then Continue;

        if tfres.classid>High(TClassColors) then
        begin
          bmp.Canvas.Pen.Color := clBlack;
          bmp.Canvas.Brush.Color := clBlack;
        end else
        begin
          bmp.Canvas.Pen.Color := TClassColors[tfres.classid];
          bmp.Canvas.Brush.Color := TClassColors[tfres.classid];
        end;
        bmp.Canvas.Brush.Style := bsClear;
//        bmp.Canvas.Rectangle(TTFBoxResult(tfres).posrect);
        with TTFBoxResult(tfres).posrect do
            bmp.Canvas.RoundRect(Left,Top,Right,Bottom, 20,20);

        bmp.Canvas.Brush.Style:= bsSolid;
        if tfres.classid>High(TClassColors) then
        begin
          bmp.Canvas.Brush.Color := clBlack;
        end else
        begin
          bmp.Canvas.Brush.Color := TClassColors[tfres.classid];
        end;
        bmp.Canvas.Font.Color := clWhite;
        cap:= tfres.labelname + ' - '  + FloatToStr(RoundTo(tfres.score,-2));
        bmp.Canvas.TextOut(TTFBoxResult(tfres).posrect.Left + 2, TTFBoxResult(tfres).posrect.Top + 1,  cap);

        if tfres.id>0 then
        begin
          bmp.Canvas.Brush.Color:=clYellow;
          bmp.Canvas.Font.Color := clRed;
          ltxt:=bmp.Canvas.TextWidth(cap);
          bmp.Canvas.TextOut(TTFBoxResult(tfres).posrect.Left +ltxt + 5, TTFBoxResult(tfres).posrect.Top + 1,
              IntToStr(tfres.id) );
        end;
    end;

end;

procedure TTFlstResults.scale(scaleWidth: Single; scaleHeight: Single);
var
  tfres: TTFResult;
begin
      for tfres in self do
      begin
            if (tfres is TTFBoxResult) then
            begin
                TTFBoxResult(tfres).posrect.Left:=Round(scaleWidth*TTFBoxResult(tfres).posrect.Left);
                TTFBoxResult(tfres).posrect.Right:=Round(scaleWidth*TTFBoxResult(tfres).posrect.Right);

                TTFBoxResult(tfres).posrect.Top:=Round(scaleHeight*TTFBoxResult(tfres).posrect.Top);
                TTFBoxResult(tfres).posrect.Bottom:=Round(scaleHeight*TTFBoxResult(tfres).posrect.Bottom);
            end;
      end;
end;
end.
